package org.jewel.core;

import java.io.File;
import java.io.FileFilter;
import java.lang.reflect.Method;
import java.util.regex.Pattern;

import javax.servlet.ServletContext;

import org.jewel.annotations.Action;
import org.jewel.annotations.Controller;
import org.jewel.annotations.ControllerScope;
import org.jewel.exceptions.ActionAlreadyExistsException;
import org.jewel.exceptions.InvalidPathPatternException;
import org.jewel.exceptions.NoActionFoundException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>Gestor d'accions. S'encarrega de cercar les accions que el programador ha creat
 * i proporcionar mtodes per relacionar-les amb les URL de les peticions.</p>
 * 
 * @author Juanma
 * @version 1.0
 *
 */
public class ActionManager {

	private static final Logger _log = LoggerFactory.getLogger(ActionManager.class);
	
	/**
	 * Mapa d'accions
	 */
	private ActionMapper mapper;
	
	
	/**
	 * Constructor.
	 */
	public ActionManager() {
	}
	
	
	/**
	 * <p>Configura el gestor cercant les classes controladores i les seves accions
	 * als paquets de l'aplicaci.</p>
	 * @param servletContext javax.servlet.ServletContext
	 */
	public void configure(ServletContext servletContext) {
		File file = new File(servletContext.getRealPath("/WEB-INF/classes"));
		try {
			mapper = new ActionMapper();
			scanResources(file, file);
		}
		catch (Exception e) {
			_log.error("Error al configurar el gestor d'accions", e);
		}
	}

	
	/**
	 * <p>Cerca una acci a partir del patr URL</p>
	 * 
	 * @param pattern Patr de la classe Action que es cerca
	 * @return Classe Action corresponent al patr
	 * @throws NoActionFoundException No s'ha trobat cap acci amb el patr
	 */
	public ActionMap getAction(String pattern) throws NoActionFoundException {
		ActionMap action = mapper.findAction(pattern);
		if (action == null) throw new NoActionFoundException("No s'ha trobat l'acci " + pattern);
		return action;
	}
	
	
	/**
	 * <p>Cerca un patr URL a partir del nom complet de l'acci</p>
	 * 
	 * @param className Nom de la classe controladora
	 * @param methodName Nom del mtode d'acci
	 * @return patr URL
	 */
	public String getPattern(String className, String methodName) {
		return mapper.findPattern(className+"."+methodName);
	}
	
	
	/**
	 * <p>Comproba que el nom d'un controlador pertanyi a un controlador de l'usuari
	 * en l'mbit indicat</p>
	 * 
	 * @param controllerName Nom del controlador cercat
	 * @param scope mbit on es cerca el controlador
	 * @return Class del controlador cercat si existeix. <i>null</i> si no existeix.
	 */
	public Class<?> getControllerClass(String controllerName, ControllerScope scope) {
		Class<?> controlador = mapper.getController(controllerName);
		if (controlador != null) {
			Controller annotation = controlador.getAnnotation(Controller.class);
			if (scope == annotation.scope()) {
				return controlador;
			}
		}
		return null;
	}
	
	
	// 
	private static final String _CLASS_SUFFIX = ".class";
	private static final int _CLASS_SUFFIX_LENGTH = _CLASS_SUFFIX.length();
	private static final String _ESCAPED_PATH_SEPARATOR = Pattern.quote(File.separator);
	private static final FileFilter _FILTER = new FileFilter() {

		public boolean accept(final File file) {
			return file.isDirectory() || file.getPath().endsWith(_CLASS_SUFFIX);
		} // accept

	};
    
	/**
	 * Cerca les accions del desenvolupador
	 * 
	 * @param path
	 * @param rootPath
	 */
	private void scanResources(final File path, final File rootPath) {
		final File[] files = path.listFiles(_FILTER);
		if (null == files)
			return;
		final int sourcePathLength = rootPath.getPath().length() + 1;

		for (final File file : files) {
			if (file.isDirectory()) {
				scanResources(file, rootPath);
			} else {
				final String className = file.getPath().substring(sourcePathLength)
						.replaceAll(_ESCAPED_PATH_SEPARATOR, ".");

				try {
					final Class<?> clazz = Class.forName(className.substring(0, className.length() - _CLASS_SUFFIX_LENGTH));

					if (clazz.isAnnotationPresent(Controller.class)) {
						for (Method method : clazz.getMethods()) {
							if (method.isAnnotationPresent(Action.class)) {
								Controller cAnnotation = clazz.getAnnotation(Controller.class);
								Action mAnnotation = method.getAnnotation(Action.class);
								try {
									mapper.addAction(
											((Controller) cAnnotation).path(),
											((Action) mAnnotation).path(),
											((Controller) cAnnotation).name(),
											clazz, method);
									_log.info("Added action: "
											+ method.getName()
											+ " in controller " + className
											+ ". Path: "
											+ ((Controller) cAnnotation).path()
											+ ((Action) mAnnotation).path());
								}
								catch (InvalidPathPatternException e) {
									_log.error(e.getMessage());
								} catch (ActionAlreadyExistsException e) {
									_log.error(e.getMessage());
								}
							}
						}
					}
				}
				catch (ClassNotFoundException e) {
					_log.error("ActionManager.scanResources ERROR", e);
				}
				catch (NoClassDefFoundError e) {
					_log.error("ActionManager.scanResources ERROR", e);
				}
			}
		}
	}
	
}
