/**
 * 
 */
package org.jewel.core;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.jewel.annotations.Controller;
import org.jewel.exceptions.ActionAlreadyExistsException;
import org.jewel.exceptions.InvalidPathPatternException;
import org.jewel.util.PatternUtil;

/**
 * <p>Classe que gestiona el mapa d'accions {@link ActionMap} del gestor d'accions o {@link ActionManager}</p>
 * 
 * @author Juanma
 * @version 1.0
 *
 */
public class ActionMapper {

	/**
	 * Guarda una relaci "patr URL" - Action
	 */
	private Map<String, ActionMap> actions;
	
	/**
	 * Guarda una relaci "Controlador.accio" - "patr URL"
	 */
	private Map<String, String> patterns;
	
	/**
	 * Constructor
	 */
	public ActionMapper() {
		actions = new HashMap<String, ActionMap>();
		patterns = new HashMap<String, String>();
	}
	
	
	/**
	 * <p>Afegeix una acci al mapa d'accions.</p>
	 * <p>Valida que el patr URL de l'acci sigui vlida abans d'afegir-la</p>
	 * 
	 * @param controllerPath Direcci URL del controlador
	 * @param actionPath Direcci URL de l'acci
	 * @param nameController Nom del controlador (atribut <i>name</i> de l'anotaci {@link Controller}</p>
	 * @param classe Classe controladora
	 * @param method Mtode que representa l'acci
	 * @throws InvalidPathPatternException Error al patr URL de l'acci
	 * @throws ActionAlreadyExistsException 
	 */
	public void addAction(String controllerPath, String actionPath, String nameController, Class<?> classe, Method method) throws InvalidPathPatternException, ActionAlreadyExistsException {
		String completePath = controllerPath + actionPath;
		if (!PatternUtil.validateActionPattern(completePath)) {
			throw new InvalidPathPatternException("El controlador " + classe.getName() + " no t un patr URL correcte a l'acci " + method.getName() + ": '" + completePath + "'");
		}
		String actionName = nameController+"."+method.getName();
		
		// Comprobem que una acci no existeixi
		if (actions.containsKey(completePath)) {
			throw new InvalidPathPatternException("El patr URL " + completePath + " ja existeix.");
		}
		if (patterns.containsKey(actionName)) {
			throw new ActionAlreadyExistsException("La acci " + actionName + " ja existeix.");
		}
		
		ActionMap action = new ActionMap(nameController, classe, method);
		actions.put(completePath, action);
		patterns.put(actionName, controllerPath + actionPath);
	}
	
	
	/**
	 * <p>Retorna l'acci a partir d'un patr URL</p>
	 * 
	 * @param pattern
	 * @return Acci que coincideix amb el patr
	 */
	public ActionMap findAction(String pattern) {
		for (String key : actions.keySet()) {
			if (PatternUtil.urlMatches(key, pattern)) return actions.get(key);
		}
		return actions.get(pattern);
	}
	
	
	/**
	 * <p>Retorna un patr a partir del nom complet de l'acci</p>
	 * <p>S'utilitza per a que es pugui obtenir la URL d'una acci a partir d'una 
	 * expressi del tipus '#{Controlador.nomAcci }'</p> 
	 * 
	 * @param actionName Nom de la l'acci amb el patr: '<classe controladora>.<nom del mtode de l'acci>'
	 * @return patr URL de l'acci
	 */
	public String findPattern(String actionName) {
		return patterns.get(actionName);
	}
	
	
	/**
	 * Cerca un controlador pel nom donat
	 * 
	 * @param controllerName Nom del controlador a cercar
	 * @return Class del controlador o <i>null</i> si no existeix.
	 */
	public Class<?> getController(String controllerName) {
		for (Entry<String, ActionMap> action : actions.entrySet()) {
			if (action.getValue().getControllerName().equals(controllerName)) {
				return action.getValue().getController();
			}
		}
		return null;
	}
	
}
