package org.jewel.core;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.digester.Digester;
import org.xml.sax.SAXException;

/**
 * <p>Configuraci de l's de Jewel a l'aplicaci.</p>
 * <p>Segueix el patr Singleton, pel que noms hi ha una instancia de la classe
 * a tota l'aplicaci</p>
 * 
 * @author Juanma
 * @version 1.0
 *
 */
public class ApplicationConfiguration {

	/**
	 * Instancia de la classe
	 */
	private static ApplicationConfiguration _instance;
	
	/**
	 * Mode de depuraci actiu
	 */
	private boolean debug;
	
	/**
	 * Configuraci de la internacionalitzaci
	 */
	private LocaleConfiguration localeConfig;
	
	/**
	 * Filtre de validaci de mostra actiu
	 */
	private boolean tokenFilter;
	
	/**
	 * Filtre de internacionalitzaci actiu
	 */
	private boolean localeFilter;
	
	/**
	 * Auto-escanejar els filtres definits pel programador
	 */
	private boolean autoscanFilters;
	
	/**
	 * <p>Llistat de filtres del programador ordenats per ordre d'execuci.</p>
	 * <p>Aquests filtres no s'utilitzaran en cas que l'atribut 'autoscanFilters' sigui <i>true</i></p>
	 */
	private List<String> customFilters;
	
	/**
	 * Pgina d'error definida per l'usuari
	 */
	private String errorPage;
	
	/**
	 * Tipus de la pgina d'error definida per l'usuari (TILES o JSP)
	 */
	private String errorPageType;
	
	/**
	 * Nom del Servlet JewelServlet definit al <i>web.xml</i>.
	 */
	private String servletName;
	
	/**
	 * <p>Patr URL de JewelServlet definit al <i>web.xml</i>.</p>
	 * <p>Necessari per poder formar b les URL de les accions</p>
	 */
	private String servletUrlPattern;
	
	
	/**
	 * Valors per defecte de la configuraci
	 */
	private static final String DEFAULT_MESSAGES_VARIABLE = "msgs";
	private static final String DEFAULT_CHANGE_PARAMETER = "language";
	
	
	public static ApplicationConfiguration getInstance() {
		if (_instance == null) {
			_instance = new ApplicationConfiguration();
		}
		return _instance;
	}
	
	/**
	 * <p>Constructor privat a un Singleton</p>
	 */
	private ApplicationConfiguration() {
		this.localeConfig = new LocaleConfiguration();
		this.customFilters = new ArrayList<String>();
	}
	
	@Override
	protected Object clone() throws CloneNotSupportedException {
		// Els Singleton no es poden clonar
		throw new CloneNotSupportedException();
	}
	
	
	/**
	 * <p>Configuram l'aplicaci a partir dels arxius <i>jewel-config.xml</i> i <i>web.xml</i></p>
	 * 
	 * @param configFile Arxiu de configuraci del framework <i>jewel-config.xml</i>
	 * @param webFile Arxiu de configuraci de l'aplicaci J2EE <i>web.xml</i>
	 * @throws IOException Error al parsejar un arxiu de configuraci
	 * @throws SAXException Error al parsejar un arxiu de configuraci
	 */
	public void configure(InputStream configFile, InputStream webFile) throws IOException, SAXException {
		// Extreiem la configuraci de l'arxiu jewel-config
        Digester digester = new Digester();
        digester.push(this);
        digester.setNamespaceAware(true);
        digester.setValidating(false);

        // Processam l'arxiu de configuraci de jewel (p.defecte: jewel-config.xml)
        digester.addBeanPropertySetter("jewel-config/debug-mode", "debug");
        digester.addBeanPropertySetter("jewel-config/jewel-filters/token-filter", "tokenFilter");
        digester.addBeanPropertySetter("jewel-config/jewel-filters/locale-filter", "localeFilter");
        digester.addBeanPropertySetter("jewel-config/error-page/location", "errorPage");
        digester.addBeanPropertySetter("jewel-config/error-page/type", "errorPageType");
        digester.addBeanPropertySetter("jewel-config/custom-filters/autoscan", "autoscanFilters");
        digester.addCallMethod("jewel-config/custom-filters/class", "addCustomFilter", 1);
        digester.addCallParam("jewel-config/custom-filters/class", 0);
        // Propietats de internacionalitzaci de l'arxiu de configuraci
        digester.addObjectCreate("jewel-config/locale-config", LocaleConfiguration.class);
        digester.addCallMethod("jewel-config/locale-config/default-language", "addDefaultLanguage", 1);
        digester.addCallParam("jewel-config/locale-config/default-language", 0);
        digester.addBeanPropertySetter("jewel-config/locale-config/messages-file", "messagesFile");
        digester.addBeanPropertySetter("jewel-config/locale-config/change-locale-parameter", "changeParameter");
        digester.addBeanPropertySetter("jewel-config/locale-config/resource-bundle-variable", "messagesVariable");
        digester.addCallMethod("jewel-config/locale-config/language", "addLanguage", 1);
        digester.addCallParam("jewel-config/locale-config/language", 0);
        digester.addSetNext("jewel-config/locale-config", "setLocaleConfig");
        
        digester.parse(configFile);
        
        // Processar l'arxiu web.xml
        digester = new Digester();
        digester.push(this);
        digester.setNamespaceAware(true);
        digester.setValidating(false);

        digester.addBeanPropertySetter("web-app/servlet-mapping/servlet-name", "servletName");
        digester.addBeanPropertySetter("web-app/servlet-mapping/url-pattern", "servletUrlPattern");
       
       	digester.parse(webFile);
       	
       	if (this.isLocaleFilter()) {
       		// Validem la configuraci de la localitzaci
       		if (this.localeConfig == null)
       			this.localeConfig = new LocaleConfiguration();
       		if (this.localeConfig.getMessagesVariable() == null) 
       			this.localeConfig.setMessagesVariable(DEFAULT_MESSAGES_VARIABLE);
       		if (this.localeConfig.getChangeParameter() == null)
       			this.localeConfig.setChangeParameter(DEFAULT_CHANGE_PARAMETER);
       	}
	}

	
	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer("********* Jewel Configuration *********\n");
		sb.append("Debug: " + this.debug + "\n");
		sb.append("TokenFilter: "+ this.tokenFilter + "\n");
		sb.append("ServletName: " + this.servletName + "\n");
		sb.append("ServletPattern: " + this.servletUrlPattern + "\n");
		sb.append("ErrorPage: " + this.errorPage + "\n");
		sb.append("ErrorPage type: " + this.errorPageType + "\n");
		sb.append("\nLocaleFilter: " + this.localeFilter + "\n");
		sb.append(this.localeConfig + "\n");
		return sb.toString();
	}

	
	// Setters / Getters
	//
	public boolean isDebug() {
		return debug;
	}
	public void setDebug(boolean debug) {
		this.debug = debug;
	}
	public LocaleConfiguration getLocaleConfig() {
		return localeConfig;
	}
	public void setLocaleConfig(LocaleConfiguration localeConfig) {
		this.localeConfig = localeConfig;
	}
	public boolean isTokenFilter() {
		return tokenFilter;
	}
	public void setTokenFilter(boolean tokenFilter) {
		this.tokenFilter = tokenFilter;
	}
	public boolean isLocaleFilter() {
		return localeFilter;
	}
	public void setLocaleFilter(boolean localeFilter) {
		this.localeFilter = localeFilter;
	}
	public boolean isAutoscanFilters() {
		return autoscanFilters;
	}
	public void setAutoscanFilters(boolean autoscanFilters) {
		this.autoscanFilters = autoscanFilters;
	}
	public List<String> getCustomFilters() {
		return customFilters;
	}
	public void addCustomFilter(String customFilter) {
		this.customFilters.add(customFilter);
	}
	public String getServletName() {
		return servletName;
	}
	public void setServletName(String servletName) {
		this.servletName = servletName;
	}
	public String getServletUrlPattern() {
		return servletUrlPattern;
	}
	public void setServletUrlPattern(String servletUrlPattern) {
		this.servletUrlPattern = servletUrlPattern;
	}
	public String getErrorPage() {
		return errorPage;
	}
	public void setErrorPage(String errorPage) {
		this.errorPage = errorPage;
	}
	public String getErrorPageType() {
		return errorPageType;
	}
	public void setErrorPageType(String errorPageType) {
		this.errorPageType = errorPageType;
	}
	
}
