/**
 * 
 */
package org.jewel.core;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.apache.commons.beanutils.BeanUtils;
import org.jewel.annotations.Action;
import org.jewel.annotations.ControllerScope;
import org.jewel.annotations.RolesAllowed;
import org.jewel.annotations.RolesDenied;
import org.jewel.core.dispatchers.JspResultDispatcher;
import org.jewel.core.dispatchers.TilesResultDispatcher;
import org.jewel.exceptions.ActionDeniedException;
import org.jewel.exceptions.InvalidParameterActionException;
import org.jewel.exceptions.NoActionFoundException;
import org.jewel.exceptions.ParameterCastException;
import org.jewel.exceptions.ResultDispatchException;
import org.jewel.exceptions.TokenException;
import org.jewel.exceptions.UndefinedActionException;
import org.jewel.util.ParameterUtil;
import org.jewel.util.PatternUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>Singleton que proporciona el control de l'aplicaci:
 * <ul><li>configuraci d'accions</li>
 * <li>execuci d'accions</li>
 * <li>configuraci de vistes</li></ul>
 * </p>
 * 
 * @author Juanma
 * @version 1.0
 *
 */
public class ApplicationController {

	private static final Logger _log = LoggerFactory.getLogger(ApplicationController.class);
	
	/**
	 * Instancia nica de la classe ApplicationController
	 */
	private static ApplicationController _instance;
	
	/**
	 * Gestor d'accions
	 */
	private ActionManager actionMgr;
	
	/**
	 * Gestor de vistes
	 */
	private ViewManager viewMgr;
	
	/**
	 * Gestor de filtres
	 */
	private FilterManager filterMgr;
	
	/**
	 * Configuraci Jewel de l'aplicaci
	 */
	private ApplicationConfiguration config;
	
	/**
	 * Constructor privat.
	 */
	private ApplicationController() {
	}
	
	
	/**
	 * <p>Obt la instancia d'aquesta classe (Singleton) dins l'aplicaci.</p>
	 * 
	 * @return la instancia de la classe a l'aplicaci.
	 */
	public static synchronized ApplicationController getInstance() {
		if (_instance == null) {
			_instance = new ApplicationController();
		}
		return _instance;
	}
	
	
	@Override
	protected Object clone() throws CloneNotSupportedException {
		// Els Singleton no es poden clonar
		throw new CloneNotSupportedException();
	}
	
	
	/**
	 * <p>Configura l'aplicaci a partir de l'arxiu <i>web.xml</i> i l'arxiu <i>jewel-config.xml</i>.</p>
	 * <p>S'encarrega de configurar el framework, les accions, les vistes i els filtres)</p>
	 * 
	 * @param servletContext Contexte del Servlet
	 * @param configFilename Nom de l'arxiu de configuraci de Jewel
	 * @throws ServletException Error de configuraci del framework
	 */
	public void configure(ServletContext servletContext, String configFilename) throws ServletException {
		config = ApplicationConfiguration.getInstance();
		
		// Cercam l'arxiu de configuraci de l'aplicaci web.xml
		InputStream webFile = servletContext.getResourceAsStream("/WEB-INF/web.xml");
		if (webFile == null) {
            throw new ServletException("No s'ha trobat l'arxiu web.xml");
        }
		
		// Cercam l'arxiu de configuraci del framework jewel-config.xml
		InputStream configFile = servletContext.getResourceAsStream(configFilename);
		if (configFile == null) {
	        throw new ServletException("No s'ha trobat l'arxiu de configuraci del framework. Ha de crear l'arxiu 'jewel-config.xml' dins " +
	        		"la carpeta WEB-INF o b especificar on es troba l'arxiu al parmetre inicial del servlet 'org.jewel.CONFIG_FILE'.");
	    }
		
		// Configurem el framework
		try {
			config.configure(configFile, webFile);
		}
		catch (Exception e) {
        	throw new ServletException(e);
		}
        finally {
        	try {
        		configFile.close();
			} catch (IOException e) {
				throw new ServletException(e);
			}
        	try {
        		webFile.close();
        	} catch (IOException e) {
				throw new ServletException(e);
			}
        }
        
        _log.info(this.config.toString());
        
		// Configurem el gestor d'accions
		actionMgr = new ActionManager();
		actionMgr.configure(servletContext);
		
		// Configurem el gestor de vistes
		viewMgr = new ViewManager();
		viewMgr.configure(servletContext);
		
		// Configurem el gestor de filtres 
		filterMgr = new FilterManager();
		filterMgr.configure(servletContext);
	}
	

	/**
	 * <p>Redirigeix a una vista determinada al {@link ModelAndView}</p>
	 * 
	 * @param mav ModelAndView que cont el Model i les dades per presentar la vista adequada.
	 * @param req ServletRequest
	 * @param resp ServletResponse
	 * @throws ResultDispatchException Error al redirigir a la vista.
	 */
	public void dispatch(ModelAndView mav, ServletRequest req, ServletResponse resp) throws ResultDispatchException {
		if (mav == null) throw new ResultDispatchException("No s'ha trobat cap model ni vista", new NullPointerException("ModelAndView"));
		
		ResultDispatcher resultDispatcher = viewMgr.getResultDispatcher(mav.getResultType());
		
		resultDispatcher.dispatch(mav, req, resp);
	}
	
	
	/**
	 * <p>Redirigeix la petici a una pgina d'error definida per l'usuari o, en cas contrari,
	 * a una pgina que crea Jewel on es mostra l'error.</p>
	 * 
	 * @param e Excepci produda
	 * @param req ServletRequest
	 * @param resp ServletResponse
	 * @throws ResultDispatchException Error al redirigir a la vista.
	 */
	public void dispatchErrorPage(Exception e, ServletRequest req, ServletResponse resp) throws ResultDispatchException {
		ResultDispatcher resultDispatcher = null;
		String errorPage = null;
		if (!config.isDebug()) {
			if ("TILES".equals(config.getErrorPageType())) {
				// Pagina d'error TILES definida per l'usuari
				resultDispatcher = viewMgr.getResultDispatcher(TilesResultDispatcher.class);
				errorPage = config.getErrorPage();
			}
			else if ("JSP".equals(config.getErrorPageType())){
				// Pagina d'error JSP definida per l'usuari
				resultDispatcher = viewMgr.getResultDispatcher(JspResultDispatcher.class);
				errorPage = config.getErrorPage();
			}
			else {
				// Pagina d'error creada per Jewel
				resultDispatcher = viewMgr.getResultDispatcher(JspResultDispatcher.class);
			}
			resultDispatcher.dispatchError(e, errorPage, req, resp);
		}
		else {
			// Pagina d'error amb informaci per la depuraci
			// 
			try {
				resp.setContentType("text/html");
				resp.getWriter().println("<html><head><title>Jewel Error Page</title></head><body>");
				resp.getWriter().println("<h1>Error</h1>");
				resp.getWriter().println("<p>" + e.getMessage() + "</p>");
				resp.getWriter().println("<h2>StackTrace</h2>");
				
				Throwable cause = e;
				while (cause != null) {
					resp.getWriter().println("<p style=\"margin-left:25px;\">");
					for (StackTraceElement st : cause.getStackTrace()) {
						resp.getWriter().println(st.toString() + "<br />");
					}
					resp.getWriter().println("</p>");
					cause = cause.getCause();
					if (cause != null) {
						resp.getWriter().println("<p>Caused by:");
						if (cause.getMessage() != null) {
							resp.getWriter().println("<br />" + cause.getMessage());
						}
						resp.getWriter().println("</p>");
					}
					
				}
				resp.getWriter().println("<hr/>");
				resp.getWriter().println("<p>Pot deshabilitar aquesta pgina canviant la configuraci del 'debug-mode' a l'arxiu de configuraci de Jewel.</p>");
				resp.getWriter().println("</body></html>");
				resp.getWriter().close();
			} catch (IOException ex) {
				throw new ResultDispatchException("No s'ha pogut presentar la vista per l'error " + e, ex);
			}
		}
		
	}

	
	/**
	 * <p>Executa una acci d'un controlador a partir d'un patr URL</p>
	 * 
	 * @param ctx Context de la petici
	 * @throws NoActionFoundException No s'ha trobat cap acci que coincideixi amb la URL
	 * @throws InvalidParameterActionException No coincideix el nmero o el tipus dels parmetres
	 * @throws UndefinedActionException S'ha produt un error al executar l'acci
	 * @throws ActionDeniedException L'usuari no t perms per executar l'acci
	 * @return ModelAndView Model i vista retornat per l'acci
	 * @throws TokenException Excepci llanada per {@link TokenFilter} si no es valida la mostra o token
	 * @throws ParameterCastException Error al convertir els parmetres als tipus de l'acci
	 */
	public ModelAndView execute(RequestContext ctx) throws NoActionFoundException, InvalidParameterActionException, UndefinedActionException, ActionDeniedException, TokenException, ParameterCastException {
		// Cercam l'acci a executar
		//
		String urlPattern = this.config.getServletUrlPattern().replace("/*", "");
		urlPattern = urlPattern.replace("*", "");
		String sAction = ctx.getUrl().replace(urlPattern, "");
		
		ActionMap action = actionMgr.getAction(sAction);
		
		//Executar els filtres pre-acci
		//
		filterMgr.executePreFilters(ctx, action);
		
		//Comprobam que l'usuari t permisos d'execuci
		//
		boolean denied = false;
		if (action.getMethod().isAnnotationPresent(RolesAllowed.class)) {
			denied = true;
			for (String role : action.getMethod().getAnnotation(RolesAllowed.class).roles()) {
				if (ctx.getRequest().isUserInRole(role)) {
					denied = false;
					break;
				}
			}
		}
		if (action.getMethod().isAnnotationPresent(RolesDenied.class)) {
			for (String role : action.getMethod().getAnnotation(RolesDenied.class).roles()) {
				if (ctx.getRequest().isUserInRole(role)) {
					denied = true;
					break;
				}
			}
		}
		if (denied) throw new ActionDeniedException("L'acci " + sAction + " no est permesa per l'usuari");
		
		Object controlador = null;
		// Instanciam la classe controladora de l'acci
		//
		if (action.isSessionScope() && ctx.getSessionControllers().get(action.getControllerName()) != null) {
			// Es tracta d'un controlador de sessi ja instanciat, pel que hem de recuperar-lo
			controlador = ctx.getSessionControllers().get(action.getControllerName());
			// Posam els parmetres del controlador que s'hagin enviat al formulari
			//
			Map<String, Object> controllerParams = ctx.getControllerParameters(action.getControllerName());
			for (Entry<String, Object> param : controllerParams.entrySet()) {
				if (param.getKey().contains(".")) {
					// Noms posam valor als objectes del controlador, 
					String beanName = param.getKey().substring(0, param.getKey().indexOf("."));
					String propertyName = param.getKey().substring(param.getKey().indexOf(".")+1);
					
					try {
						Object bean = BeanUtils.getProperty(controlador, beanName);
						if (bean != null) {
								bean = ParameterUtil.getProperty(bean, propertyName);
								BeanUtils.setProperty(bean, propertyName, param.getValue());
						} else {
							_log.error("No es pot escriure la propietat '" + param.getKey() + "' del controlador '" + action.getControllerName() + "'");
						}
					}
					catch (Exception e) {
						_log.error("No es pot escriure la propietat '" + param.getKey() + "' del controlador '" + action.getControllerName() + "'", e);
					}
				}
				else {
					// Posem valor a una propietat simple del controlador
					try {
						BeanUtils.setProperty(controlador, param.getKey(), param.getValue());
					}
					catch (Exception e) {
						_log.error("No es pot escriure la propietat '" + param.getKey() + "' del controlador '" + action.getControllerName() + "'", e);
					}
				}
			}
		}
		else {
			Class<?> controllerClass;
			try {
				controllerClass = Class.forName(action.getController().getName());
				controlador = controllerClass.getConstructor().newInstance();
				
				if (action.isSessionScope()) {
					// Guardem el controlador de sessi per poder-lo utilitzar en futures peticions
					ctx.addSessionController(action.getControllerName(), controlador);
				}
			}
			catch (Exception e) {
				throw new NoActionFoundException("No s'ha pogut instanciar el controlador '" + action.getControllerName()+"'", e);
			}
		}
		
		// Obtenim els parmetres d'execuci de l'acci
		//
		Object[] methodParams = getActionArguments(ctx, sAction, action);
		
		// Executam l'acci
		//
		ModelAndView mav = null;
		try {
			Object o = action.getMethod().invoke(controlador, methodParams);
			if (o instanceof ModelAndView) {
				mav = (ModelAndView) o;
			}
		}
		catch (Exception e) {
			throw new UndefinedActionException("S'ha produt un error al invocar l'acci " + sAction, e);
		}
		
		//Executar els filtres post-acci
		//
		filterMgr.executePostFilters(ctx, action);
		
		// Retornem el model i vista generats per l'acci
		return mav;
	}
	
	
	/**
	 * <p>Obt el patr URL d'una acci identificada per l'expressi 'controlador.mtode'</p>
	 * 
	 * @param expression Expressi 'controlador.mtode'
	 * @return Patr URL de l'acci
	 */
	public String getPattern(String expression) {
		if (!PatternUtil.isJewelExpression(expression)) return null;
		String pattern = null;
		
		String completeName = PatternUtil.getExpression(expression);
		if (completeName != null && completeName.contains(".")) {
			String className = completeName.substring(0, completeName.indexOf("."));
			String methodName = completeName.substring(completeName.indexOf(".")+1, completeName.length());
			
			pattern = actionMgr.getPattern(className, methodName);
		}
		
		return pattern;
	}

	
	/**
	 * <p>Patr URL del {@link JewelServlet}</p>
	 * 
	 * @return Patr URL del JewelServlet
	 */
	public String getServletUrlPattern() {
		return this.config.getServletUrlPattern();
	}
	

	/**
	 * <p>Valida que una acci est mapejada a l'aplicaci.</p>
	 * 
	 * @param url URL de l'acci
	 * @return acci mapejada
	 */
	public boolean validateAction(String url) {
		try {
			actionMgr.getAction(url);
		} catch (NoActionFoundException e) {
			return false;
		}
		return true;
	}
	
	
	/**
	 * Comproba si un nom de controlador pertany a un controlador de l'usuari i el retorna.
	 * 
	 * @param controllerName Nom del controlador
	 * @return <i>true</i> si existeix. <i>false</i> en cas contrari.
	 */
	public Class<?> getSessionControllerClass(String controllerName) {
		Class<?> controller = actionMgr.getControllerClass(controllerName, ControllerScope.SESSION);
		return controller;
	}
	
	
	private Object[] getActionArguments(RequestContext ctx, String sAction,
			ActionMap action) throws InvalidParameterActionException, ParameterCastException {
		// Tipus dels arguments de l'acci
		Class<?>[] actionTypes = action.getMethod().getParameterTypes();
		// Valors dels parametres
		Object[] methodParams = new Object[actionTypes.length]; // Anirem guardant els valors dels parmetres
		
		// Parametres declarats a la URL. Esquivem el primer '/' que s el path del controlador
		Action actionAnnotation = action.getMethod().getAnnotation(Action.class);
		Map<String, String> paramsDeclared = PatternUtil.getExpressionsUrl(actionAnnotation.path(), sAction.substring(sAction.indexOf('/', 1))); 
		int cont = 0;
		
		// Parametres GET (de la URL)
		List<String> expressions = PatternUtil.getExpressions(actionAnnotation.path());
		for (String s : expressions) {
			Class<?> paramType = actionTypes[cont];
			Object param = null;
			if (paramType.equals(Integer.class) || paramType.equals(String.class)) {
				param = ParameterUtil.getParameterObject(paramType, paramsDeclared.get(s));
				methodParams[cont++] = param;
			}
			else {
				throw new InvalidParameterActionException("El parmetre " + cont + " del mtode " + action.getController().getCanonicalName() +"."+action.getMethod().getName()
						+ " s de tipus " + paramType.getCanonicalName() + ", per a la URL noms es poden passar parmetres Integer o String.");
			}
		}
		
		// Comprobem que s'hagin especificat els parmetres que toquen
		// i que els parmetres siguin del tipus declarat al mtode
		for (String paramName : actionAnnotation.params()) { // Parametres declarats a l'anotaci
			Class<?> paramType = actionTypes[cont];
			Object paramValue = null;
			
			Object param = null;
			try {
				// Es un objecte de Java (String, Integer, Long, Float, Double, Date o Array)
				if (ParameterUtil.isJavaObject(paramType)) {
					paramValue = ctx.getParameter(paramName);
					param = ParameterUtil.getParameterObject(paramType, paramValue);
				}
				else {
					// No s un objecte de Java, deu ser un JavaBean
					param = ParameterUtil.getParameterJavabean(paramType, paramName, ctx);
				}
			}
			catch (Exception e) {
				e.printStackTrace();
				throw new InvalidParameterActionException("El parmetre " + paramName + " del mtode " + action.getController().getCanonicalName() +"."+action.getMethod().getName()
						+ " s de tipus " + paramType.getCanonicalName() + " per s'ha rebut el valor " + paramValue + " de tipus " + (paramValue != null ? paramValue.getClass().getCanonicalName() : "undefined"));
			}
			
			methodParams[cont++] = param;
		}
		return methodParams;
	}

}
