/**
 * 
 */
package org.jewel.core;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.regex.Pattern;

import javax.servlet.ServletContext;

import org.jewel.annotations.Filter;
import org.jewel.annotations.FilterAction;
import org.jewel.annotations.PostAction;
import org.jewel.annotations.PreAction;
import org.jewel.exceptions.TokenException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>Gestor de filtres. S'encarrega de cercar els filtres que el programador ha creat
 * i gestionar-los junt amb els proporcionats per Jewel.</p>
 * 
 * @author Juanma
 * @version 1.0
 *
 */
public class FilterManager {

	private static final Logger _log = LoggerFactory.getLogger(FilterManager.class);
	
	/**
	 * Llistat de filtres a executar abans de l'acci
	 */
	private List<FilterMap> preFilters;
	
	/**
	 * Llistat de filtres a executar desprs de l'acci
	 */
	private List<FilterMap> postFilters;
	
	
	/**
	 * Constructor
	 */
	public FilterManager() {
		preFilters = new ArrayList<FilterMap>();
		postFilters = new ArrayList<FilterMap>();
	}
	
	
	/**
	 * <p>Configura la classe cercant les classes filtre als paquets de l'aplicaci
	 * i afegint els filtres especificats a la configuraci ({@link TokenFilter} i/o {@link LocaleFilter})</p>
	 * @param servletContext ServletContext
	 */
	public void configure(ServletContext servletContext) {
		ApplicationConfiguration config = ApplicationConfiguration.getInstance();
		try {
			if (config.isTokenFilter()) {
				try {
					preFilters.add(new FilterMap(TokenFilter.class, TokenFilter.class.getMethod("preAction", RequestContext.class)));
					postFilters.add(new FilterMap(TokenFilter.class, TokenFilter.class.getMethod("postAction", RequestContext.class)));
					_log.info("Filtre afegit: " + TokenFilter.class.getCanonicalName());
				} catch (SecurityException e) {
					e.printStackTrace();
				} catch (NoSuchMethodException e) {
					e.printStackTrace();
				}
			}
			
			if (config.isLocaleFilter()) {
				try {
					preFilters.add(new FilterMap(LocaleFilter.class, LocaleFilter.class.getMethod("preAction", RequestContext.class)));
					_log.info("Filtre afegit: " + LocaleFilter.class.getCanonicalName());
				} catch (SecurityException e) {
					e.printStackTrace();
				} catch (NoSuchMethodException e) {
					e.printStackTrace();
				}
			}
			
			if (config.isAutoscanFilters()) {
				// Cercam els filtres del desenvolupador
				File file = new File(servletContext.getRealPath("/WEB-INF/classes"));
				scanResources(file, file);
			}
			else {
				ClassLoader ldr = Thread.currentThread().getContextClassLoader();
				// Utilitzem els filtres declarats a la configuraci
				for (String filterClass : config.getCustomFilters()) {
					try {
						Class<?> c = ldr.loadClass(filterClass);
						analizeFilterClass(c);
					}
					catch (ClassNotFoundException ex) {
						_log.error("Error carregant el filtre personal: " + filterClass, ex);
					}
				}
			}
			
			// Invertim l'ordre dels filtres post-acci ja que s'executen en ordre invers als pre-acci 
			Collections.reverse(postFilters);
		}
		catch (Exception e) {
			_log.error("Error al configurar el gestor de filtres", e);
		}
	}
	
	
	/**
	 * Executa els filtres que s'executen abans de l'acci
	 * 
	 * @param ctx Context de la petici
	 * @param action Acci
	 * @throws TokenException Excepci llanada per {@link TokenFilter} si no es valida la mostra o token
	 */
	public void executePreFilters(RequestContext ctx, ActionMap action) throws TokenException {
		for (FilterMap filtreMap : this.preFilters) {
			executeFilter(ctx, filtreMap, action);
		}
	}
	
	
	/**
	 * Executa els filtres que s'executen desprs de l'acci
	 * 
	 * @param ctx Context de la petici
	 * @param action Acci
	 * @throws TokenException Excepci llanada per {@link TokenFilter} si no es valida la mostra o token
	 */
	public void executePostFilters(RequestContext ctx, ActionMap action) throws TokenException {
		for (FilterMap filtreMap : this.postFilters) {
			executeFilter(ctx, filtreMap, action);
		}
	}
	
	
	/**
	 * Executa un filtre, sempre i quan l'acci no estigui excluida del filtre.
	 * 
	 * @param ctx Context de la petici
	 * @param filtreMap Filtre a executar
	 * @param action Acci executada a la petici
	 * @throws TokenException Excepci llanada per {@link TokenFilter} si no es valida la mostra o token
	 */
	private void executeFilter(RequestContext ctx, FilterMap filtreMap, ActionMap action) throws TokenException {
		try {
			Class<?> filtreClass = Class.forName(filtreMap.getFilter().getName());
			Filter filterAnnotation = filtreClass.getAnnotation(Filter.class);
			if (filterAnnotation != null && filterAnnotation.excludedActions().length > 0) {
				// Miram si l'acci a executar est exclosa del filtre
				for (FilterAction fAction : filterAnnotation.excludedActions()) {
					if (fAction.controller().equals(action.getController())) {
						if (fAction.methods() == null || fAction.methods().length == 0) {
							// S'exclouen del filtre tots els mtodes del controlador
							return;
						}
						for (String method : fAction.methods()) {
							// Mtodes que s'exclouen del filtre
							if (method.equals(action.getMethod().getName())) return;
						}
					}
				}
			}
			if (filterAnnotation != null && filterAnnotation.onlyActions().length > 0) {
				boolean trobat = false;
				// Miram si l'acci a executar no est inclosa al filtre
				for (FilterAction fAction : filterAnnotation.onlyActions()) {
					if (fAction.controller().equals(action.getController())) {
						if (fAction.methods() == null || fAction.methods().length == 0) {
							// Poden executar el filtre tots els mtodes del controlador
							trobat = true; break;
						}
						for (String method : fAction.methods()) {
							if (method.equals(action.getMethod().getName())) {
								// Mtodes que permesos del filtre
								trobat = true; break;
							}
						}
					}
				}
				if (!trobat) return;
			}
			Object filtre = filtreClass.getConstructor().newInstance();
			Object[] filtreArgs = new Object[]{ ctx };
			filtreMap.getMethod().invoke(filtre, filtreArgs);
		}
		catch (Exception e) {
			if (e.getCause() instanceof TokenException) throw ((TokenException) e.getCause());
			_log.error("No s'ha pogut executar el filtre " + filtreMap.getFilter().getCanonicalName(), e);
		}
	}


	// 
	private static final String _CLASS_SUFFIX = ".class";
	private static final int _CLASS_SUFFIX_LENGTH = _CLASS_SUFFIX.length();
	private static final String _ESCAPED_PATH_SEPARATOR = Pattern
			.quote(File.separator);
	private static final FileFilter _FILTER = new FileFilter() {

		public boolean accept(final File file) {
			return file.isDirectory() || file.getPath().endsWith(_CLASS_SUFFIX);
		} // accept

	};
	
	/**
	 * Cerca els filtres del desenvolupador
	 * 
	 * @param path Path on cercar els filtres
	 * @param rootPath Base d'on cercar els filtres
	 * @throws IOException Error al cercar els filtres
	 */
	private void scanResources(final File path, final File rootPath) throws IOException {
		final File[] files = path.listFiles(_FILTER);
		if (null == files)
			return;
		final int sourcePathLength = rootPath.getPath().length() + 1;

		for (final File file : files) {
			if (file.isDirectory()) {
				scanResources(file, rootPath);
			}
			else {
				final String className = file.getPath().substring(sourcePathLength)
						.replaceAll(_ESCAPED_PATH_SEPARATOR, ".");

				try {
					final Class<?> clazz = Class.forName(className.substring(0, className.length() - _CLASS_SUFFIX_LENGTH));

					analizeFilterClass(clazz);
				}
				catch (ClassNotFoundException e) {
					_log.error("FilterManager.scanResources ERROR", e);
				}
				catch (NoClassDefFoundError e) {
					_log.error("FilterManager.scanResources ERROR", e);
				}
			}
		}
	}


	/**
	 * Analitza una classe en cerca de l'anotaci Filter i dels mtodes PreAction i PostAction
	 * 
	 * @param clazz
	 */
	private void analizeFilterClass(Class<?> clazz) {
		if (clazz.isAnnotationPresent(Filter.class)) {
			for (Method m : clazz.getMethods()) {
				if (m.isAnnotationPresent(PreAction.class)) {
					preFilters.add(new FilterMap(clazz, m));
					_log.info("Filtre afegit: " + clazz.getCanonicalName() + " pre-action: " + m.getName());
				}
				// Un mtode pot ser PreAction i PostAction a la vegada
				if (m.isAnnotationPresent(PostAction.class)) {
					postFilters.add(new FilterMap(clazz, m));
					_log.info("Filtre afegit: " + clazz.getCanonicalName() + " post-action: " + m.getName());
				}
			}
		}
	}
	
}
