/**
 * Jewel Framework (Java Easy Web and Light Framework)
 * 
 * This work is licensed under the Creative Commons Atribucin-NoComercial-SinDerivadas 3.0 Espaa License. 
 * To view a copy of this license, visit http://creativecommons.org/licenses/by-nc-nd/3.0/es/ 
 * or send a letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.
 * 
 */
package org.jewel.core;

import java.io.IOException;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.jewel.exceptions.ResultDispatchException;
import org.jewel.util.JewelConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>Classe principal del framework que actuar de controlador central
 * i rebr totes les peticions.</p>
 * 
 * @author	Juanma
 * @version 1.0
 *
 */
public class JewelServlet extends HttpServlet {

	private static final long serialVersionUID = -6038667838332518004L;
	
	private static final Logger _log = LoggerFactory.getLogger(JewelServlet.class);
	
	
	/**
	 * <p>Inicialitza el framework cridant al {@link ApplicationController}.</p>
	 * 
	 * <p>Inicialitza el framework amb els parmetres inicials, les accions
	 * dels controladors, ...</p>
	 */
	@Override
	public void init() throws ServletException {
		super.init();
		
		String configFile = (getInitParameter(JewelConstants.INIT_PARAM_CONFIG) != null ? getInitParameter(JewelConstants.INIT_PARAM_CONFIG) : "/WEB-INF/jewel-config.xml");
		
		ApplicationController.getInstance().configure(this.getServletContext(), configFile);
	}
	
	
	/**
	 *  <p>Intercepta les peticions de l'aplicaci (GET i POST) i es comunica amb 
	 *  el {@link ApplicationController} per executar l'acci i per redirigir a la vista
	 *  adequada.</p>
	 */
	@Override
	protected void service(HttpServletRequest req, HttpServletResponse resp)
			throws ServletException, IOException {
		_log.info("Processing by JewelServlet!");
		boolean error = false;
		
		// Guardam la petici a l'objecte de context RequestContext
		//
		RequestContext requestCtx = new RequestContext(req, resp);
		
		// Executam l'acci indicada a la URL
		//
		ModelAndView mav = null;
		try {
			mav = ApplicationController.getInstance().execute(requestCtx);
		}
		catch (Exception e) {
			error = true;
			_log.error("Error al executar l'acci ", e);
			try {
				ApplicationController.getInstance().dispatchErrorPage(e, req, resp);
			} catch (ResultDispatchException e1) {
				_log.error("Error al presentar la pgina d'error", e1);
				throw new ServletException(e1);
			}
		}
		
		
		if (!error) {
			// Redirigim a la vista
			//
			try {
				if (mav != null) {
					req.getSession().setAttribute(JewelConstants.CURRENT_MAV_KEY, mav);
				}
				else {
					mav = (ModelAndView) req.getSession().getAttribute(JewelConstants.CURRENT_MAV_KEY);
					
					if (mav == null) {
						resp.sendRedirect(req.getContextPath());
					}
				}
				ApplicationController.getInstance().dispatch(mav, req, resp);
			}
			catch (ResultDispatchException e) {
				_log.error("Error al presentar la vista", e);
				try {
					ApplicationController.getInstance().dispatchErrorPage(e, req, resp);
				} catch (ResultDispatchException e1) {
					_log.error("Error al presentar la pgina d'error", e1);
					throw new ServletException(e1);
				}
			}
		}
	}

}
