package org.jewel.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.jewel.core.ApplicationConfiguration;

/**
 * Utilitats per validar i obtenir valors dels patrons, tant d'expressions com d'accions
 * 
 * @author Juanma
 * @version 1.0
 *
 */
public class PatternUtil {

	private static final String EXPRESSION_PATTERN_INI = "#{";
	private static final String EXPRESSION_PATTERN_FIN = "}";
	private static final String EXPRESSION_PATTERN = "#\\{\\s*[(a-zA-Z0-9)+\\.]+\\s*}";
	private static final String EXPRESSION_MESSAGE_PATTERN = "#\\{\\s*[(a-zA-Z0-9)+(\\[\\'(\\w\\|\\.)*'\\])*]+\\s*}";
	private static final String PATH_DECLARATION_PATTERN = "\\/([a-z0-9-]+)(\\/(([a-z0-9-]+)|((#)\\{[a-zA-Z0-9\\.]+\\})))+";
	
	/**
	 * Valida que la URL d'una acci estigui ben declarada
	 */
	public static boolean validateActionPattern(String pathDeclaration) {
		return validatePattern(PATH_DECLARATION_PATTERN, pathDeclaration);
	}
	
	private static boolean validatePattern(String pattern, String match) {
		Pattern p = Pattern.compile(pattern);
		Matcher m = p.matcher(match);
		return m.matches();
	}
	
	/**
	 * Obt les expressions d'un texte
	 * @param text
	 * @return Llistat de expressions trobades
	 */
	public static List<String> getExpressions(String text) {
		if (text == null || text.length() <= 3) return null;
		List<String> expressions = new ArrayList<String>();
		
		int posIni = text.indexOf(EXPRESSION_PATTERN_INI);
		while (posIni != -1) {
			int posFin = text.indexOf(EXPRESSION_PATTERN_FIN, posIni);
			expressions.add(text.substring(posIni+2, posFin).trim());
			posIni = text.indexOf(EXPRESSION_PATTERN_INI, posFin);
		}
		
		return expressions;
	}
	
	/**
	 * Obt el text de l'interior d'una expressi
	 * @param text Amb el patr "#{<texte>}
	 * @return nom de l'expressi
	 */
	public static String getExpression(String text) {
		if (text == null || text.trim().length() <= 3) return null;
		text = text.trim();
		if (text.startsWith("#{") && text.endsWith("}")) {
			return text.substring(2, text.length()-1).trim();
		}
		return null;
	}
	
	/**
	 * Obt el text de l'interior d'una expressi de missatge internacionalitzat
	 * @param text Amb el patr "#{<texte>}
	 * @param rb ResourceBundle utilitzat per la traducci
	 * @return Text tradut en l'idioma del ResourceBundle
	 */
	public static String getMessageExpression(String text, ResourceBundle rb) {
		if (text == null || text.trim().length() <= 3) return null;
		text = text.trim();
		String messageText = null;
		if (text.startsWith("#{") && text.endsWith("}")) {
			messageText = text.substring(2, text.length()-1).trim();
		} else return null;
		
		ApplicationConfiguration config = ApplicationConfiguration.getInstance();
		
		if (config.isLocaleFilter() && messageText.length() > (config.getLocaleConfig().getMessagesVariable().length() + 4) 
				&& messageText.startsWith(config.getLocaleConfig().getMessagesVariable()+"[")) {
		
			String propertyName = messageText.substring(messageText.indexOf("['")+2, messageText.indexOf("']"));
			return (rb != null && rb.containsKey(propertyName) ? (String) rb.getObject(propertyName) : "&lt;" + propertyName + "&gt;");
		}
		
		return null;
	}
	
	/**
	 * Obt les expressions d'un patr URL
	 * @param urlPattern patr URL
	 * @param url URL on trobar les expressions
	 * @return Conjunt d'expressions d'una URL
	 */
	public static Map<String, String> getExpressionsUrl(String urlPattern, String url) {
		if (url == null || urlPattern == null) return null;
		Map<String, String> expressions = new HashMap<String, String>();
		
		String[] patternParts = urlPattern.split("/");
		String[] urlParts = url.split("/");
		
		for (int i=urlParts.length-1, j=patternParts.length-1; i >= 0 && j >= 0; i--, j--) {
			if (isJewelExpression(patternParts[j])) {
				expressions.put(getExpression(patternParts[j]), urlParts[i]);
			}
		}
		
		return expressions;
	}

	/**
	 * <p>Mira que una URL amb parmetres coincideixi amb el patr URL</p>
	 * <p>Primer es comproba que el patr URL sigui vlid</p>
	 * @param text Patr URL
	 * @param completeText URL a comprovar
	 * @return <i>true</i> si el completeText coincideix amb el patr URL de text 
	 */
	public static boolean urlMatches(String text, String completeText) {
		if (!validateActionPattern(text)) return false;
		
		String elText = text;
		elText = elText.replaceAll(EXPRESSION_PATTERN, "[a-zA-Z0-9-]+");
		elText = elText.replaceAll("/", "\\\\/");
		
		Pattern p = Pattern.compile(elText);
		Matcher m = p.matcher(completeText);
				
		return m.matches();
	}
	
	/**
	 * Comprova que un texte sigui una expressi del tipus #{\b*[a-zA-Z0-9\.]+\b*}
	 * @param text a comprovar
	 * @return <i>true</i> si s una expressi Jewel; <i>false</> en cas contrari
	 */
	public static boolean isJewelExpression(String text) {
		if (text == null) return false;
		
		Pattern p = Pattern.compile(EXPRESSION_PATTERN);
		Matcher m = p.matcher(text);
		
		return m.matches();
	}
	
	
	/**
	 * Comprova que un text sigui una expressi de missatge internacionalitzat
	 * @param text a comprovar
	 * @return <i>true</i> si s una expressi de missatge internacionalitzat Jewel; <i>false</> en cas contrari
	 */
	public static boolean isJewelMessageExpression(String text) {
		if (text == null) return false;
		
		Pattern p = Pattern.compile(EXPRESSION_MESSAGE_PATTERN);
		Matcher m = p.matcher(text);
		
		if (!m.matches()) return false;
		
		ApplicationConfiguration config = ApplicationConfiguration.getInstance();
		
		String messageText = getExpression(text);
		if (config.isLocaleFilter() && messageText.length() > (config.getLocaleConfig().getMessagesVariable().length() + 4) 
				&& messageText.startsWith(config.getLocaleConfig().getMessagesVariable()+"[")) {
			// Es tracta d'un missatge que s'ha de cercar al ResourceBundle de l'idioma actual
			return true;
		}
		return false;
	}
}
